var __defProp = Object.defineProperty;
var __defNormalProp = (obj, key, value) => key in obj ? __defProp(obj, key, { enumerable: true, configurable: true, writable: true, value }) : obj[key] = value;
var __publicField = (obj, key, value) => {
  __defNormalProp(obj, typeof key !== "symbol" ? key + "" : key, value);
  return value;
};
var __async = (__this, __arguments, generator) => {
  return new Promise((resolve, reject) => {
    var fulfilled = (value) => {
      try {
        step(generator.next(value));
      } catch (e) {
        reject(e);
      }
    };
    var rejected = (value) => {
      try {
        step(generator.throw(value));
      } catch (e) {
        reject(e);
      }
    };
    var step = (x) => x.done ? resolve(x.value) : Promise.resolve(x.value).then(fulfilled, rejected);
    step((generator = generator.apply(__this, __arguments)).next());
  });
};

// src/binders.ts
function onUpdate(computed, cb, initial = false) {
  if (initial)
    cb(computed.get());
  return computed.observe(() => cb(computed.get()));
}
function input(node, value) {
  const subscription = onUpdate(value, (ref) => node.value = ref.get(), true);
  const listener = () => value.get().set(node.value);
  node.addEventListener("input", listener);
  return {
    destroy() {
      node.removeEventListener("input", listener);
      subscription.revoke();
    }
  };
}

// src/event-emitter.ts
var EventEmitter = class {
  constructor() {
    __publicField(this, "listeners", /* @__PURE__ */ new Map());
  }
  on(event, listener) {
    let listeners;
    if (this.listeners.has(event))
      (listeners = this.listeners.get(event)).push(listener) - 1;
    else
      this.listeners.set(event, listeners = [listener]);
    return {
      revoke: () => {
        listeners.splice(listeners.indexOf(listener), 1);
      }
    };
  }
  trigger(event, ...args) {
    if (!this.listeners.has(event))
      return;
    this.listeners.get(event).forEach((listener) => listener(...args));
  }
};

// src/modifiers.ts
var Modifiers = class {
  constructor() {
    __publicField(this, "modifiers", /* @__PURE__ */ new WeakMap());
  }
  set(target, key, modifier, value) {
    let map;
    if (this.modifiers.has(target))
      map = this.modifiers.get(target);
    else
      this.modifiers.set(target, map = /* @__PURE__ */ new Map());
    if (map.has(key))
      map.get(key)[modifier] = value;
    else
      map.set(key, { [modifier]: value });
  }
  get(target, key, modifier) {
    var _a, _b;
    return (_b = (_a = this.modifiers.get(Object.getPrototypeOf(target))) == null ? void 0 : _a.get(key)) == null ? void 0 : _b[modifier];
  }
  has(target, key, modifier) {
    return this.get(target, key, modifier) !== void 0;
  }
};

// src/scheduler.ts
var Scheduler = class {
  constructor() {
    __publicField(this, "queue", /* @__PURE__ */ new Set());
    __publicField(this, "microtask", Promise.resolve());
    __publicField(this, "flushing", false);
  }
  next(cb) {
    void this.microtask.then(cb);
  }
  flush() {
    for (const task of this.queue) {
      task();
      this.queue.delete(task);
    }
  }
  enqueue(task) {
    this.queue.add(task);
    if (!this.flushing) {
      this.next(() => {
        this.flushing = false;
        this.flush();
      });
    }
  }
};

// src/model.ts
function createLink(target, key) {
  return { target, key };
}
function isEqualLink(x, y) {
  return x.target === y.target && x.key === y.key;
}
var modifiers = new Modifiers();
function assertNotModifier(target, key, modifier, message) {
  if (modifiers.has(target, key, modifier))
    throw new Error(message);
}
function assertNotIgnored(target, key, message = `Property ${String(key)} is ignored.`) {
  assertNotModifier(target, key, "ignore", message);
}
function ignore(target, key) {
  assertNotIgnored(target, key, `Property ${String(key)} is already ignored.`);
  modifiers.set(target, key, "ignore", true);
}
function effect(target, key) {
  assertNotIgnored(target, key);
  modifiers.set(target, key, "effect", true);
}
function throttle(delay) {
  return (target, key) => {
    assertNotIgnored(target, key);
    assertNotModifier(target, key, "throttle", `Property ${String(key)} is already throttled.`);
    assertNotModifier(target, key, "debounce", `Cannot throttle property ${String(key)}. Property is already debounced.`);
    modifiers.set(target, key, "throttle", delay);
  };
}
function debounce(delay) {
  return (target, key) => {
    assertNotIgnored(target, key);
    assertNotModifier(target, key, "debounce", `Property ${String(key)} is already debounced.`);
    assertNotModifier(target, key, "throttle", `Cannot debounce property ${String(key)}. Property is already throttled.`);
    modifiers.set(target, key, "debounce", delay);
  };
}
function readonly(target, key) {
  assertNotModifier(target, key, "readonly", `Property ${String(key)} is already readonly.`);
  modifiers.set(target, key, "readonly", true);
}
function extend(extender) {
  return (target, key) => {
    if (modifiers.has(target, key, "extend"))
      modifiers.get(target, key, "extend").push(extender);
    else
      modifiers.set(target, key, "extend", [extender]);
  };
}
var Model = class {
  constructor() {
    __publicField(this, "$__events", new EventEmitter());
    __publicField(this, "$__scheduler", new Scheduler());
    const extend2 = (object, isRoot) => {
      let getRef = false;
      let silent = false;
      const getters = /* @__PURE__ */ new WeakMap();
      const computed = (target, key, receiver, descriptior) => {
        var _a;
        if (getters.has(descriptior.get))
          return getters.get(descriptior.get)();
        let value;
        let dirty = true;
        let dispose;
        let debounceTimeout;
        const observers = [];
        const extenders = (_a = modifiers.get(target, key, "extend")) != null ? _a : [];
        const get = () => {
          if (dirty) {
            const dependencies = [];
            const listener = this.$__events.on("get", (link) => dependencies.push(link));
            const next = Reflect.get(target, key, receiver);
            listener.revoke();
            dirty = false;
            if (next !== value) {
              value = next;
              for (const extender of extenders) {
                if (extender.compute)
                  value = extender.compute(value);
              }
              dispose == null ? void 0 : dispose();
              const listeners = dependencies.map((dependency) => this.$__observe(dependency, () => __async(this, null, function* () {
                dirty = true;
                for (const extender of extenders) {
                  if (extender.recompute && !(yield extender.recompute())) {
                    dirty = false;
                    return;
                  }
                  if (extender.notify && !(yield extender.notify(value)))
                    return;
                }
                const notify = () => {
                  this.$__events.trigger("set", createLink(target, key));
                  for (const listener2 of observers)
                    listener2 == null ? void 0 : listener2();
                };
                if (modifiers.has(target, key, "throttle")) {
                  const timeout = modifiers.get(target, key, "throttle");
                  setTimeout(notify, timeout);
                } else if (modifiers.has(target, key, "debounce")) {
                  const timeout = modifiers.get(target, key, "debounce");
                  if (debounceTimeout !== void 0)
                    clearTimeout(debounceTimeout);
                  debounceTimeout = setTimeout(notify, timeout);
                } else {
                  this.$schedule(notify);
                }
              })));
              dispose = () => listeners.map((l) => l.revoke());
            }
          }
          return value;
        };
        const observe = (listener) => {
          const index = observers.push(listener) - 1;
          return {
            revoke() {
              observers[index] = void 0;
            }
          };
        };
        for (const extender of extenders) {
          if (extender.initialize) {
            extender.initialize({
              get,
              observe,
              dispose() {
                throw new Error("Cannot dispose of computed property.");
              }
            });
          }
        }
        getters.set(descriptior.get, get);
        return get();
      };
      return new Proxy(object, {
        get: (target, key, receiver) => {
          const createReference = (link) => ({
            get: () => Reflect.get(link.target, link.key),
            set: (value2) => void Reflect.set(link.target, link.key, value2),
            observe: (listener) => this.$__observe(link, listener)
          });
          if (isRoot) {
            if (key === this.$ref.name) {
              getRef = true;
              return (link) => {
                return createReference(link);
              };
            }
            if (key === this.$silent.name) {
              silent = true;
              return (value2) => {
                silent = false;
                return value2;
              };
            }
            if (typeof key === "string" && key.startsWith("$") || modifiers.get(target, key, "ignore") === true)
              return Reflect.get(target, key, receiver);
          }
          if (getRef) {
            getRef = false;
            return createLink(receiver, key);
          }
          const get = () => {
            const descriptior = Object.getOwnPropertyDescriptor(Object.getPrototypeOf(target), key);
            if (descriptior == null ? void 0 : descriptior.get)
              return computed(target, key, receiver, descriptior);
            return Reflect.get(target, key, receiver);
          };
          const value = get();
          if (!silent)
            this.$__events.trigger("get", createLink(target, key));
          if (typeof value === "object" && value !== null)
            return extend2(value, false);
          return value;
        },
        set: (target, key, value, receiver) => {
          const succeeded = Reflect.set(target, key, value, receiver);
          this.$schedule(() => {
            this.$__events.trigger("set", createLink(target, key));
          });
          return succeeded;
        }
      });
    };
    const extended = extend2(this, true);
    for (const key in this.constructor.prototype) {
      if (Object.prototype.hasOwnProperty.call(this.constructor.prototype, key)) {
        if (modifiers.get(this.constructor.prototype, key, "effect") === true)
          Reflect.get(extended, key);
      }
    }
    return extended;
  }
  $__observe(property, listener) {
    return this.$__events.on("set", (link) => {
      if (isEqualLink(property, link))
        listener();
    });
  }
  $__observeDependencies(callback, listener) {
    const dependencies = [];
    const dependencyListener = this.$__events.on("get", (link) => dependencies.push(link));
    callback();
    dependencyListener.revoke();
    const listeners = dependencies.map(
      (dependency) => this.$__observe(dependency, () => {
        listener();
      })
    );
    return {
      revoke() {
        for (const listener2 of listeners)
          listener2.revoke();
      }
    };
  }
  $silent(target) {
    throw new Error("not implemented");
  }
  $ref(target) {
    throw new Error("not implemented");
  }
  $observe(target, listener) {
    return this.$__observe(target, listener);
  }
  $computed(get) {
    const subscriptions = [];
    let value;
    let dirty = true;
    let dispose;
    let disposed = false;
    const assertNotDisposed = () => {
      if (disposed)
        throw new Error("Reference has been disposed.");
    };
    return {
      get: () => {
        assertNotDisposed();
        if (dirty) {
          const dependencies = [];
          const listener = this.$__events.on("get", (link) => dependencies.push(link));
          const next = get();
          listener.revoke();
          dirty = false;
          if (next !== value) {
            value = next;
            dispose == null ? void 0 : dispose();
            const listeners = dependencies.map((dependency) => this.$__observe(dependency, () => {
              dirty = true;
              this.$__scheduler.enqueue(() => {
                subscriptions.forEach((listener2) => listener2());
              });
            }));
            dispose = () => listeners.map((l) => l.revoke());
          }
        }
        return value;
      },
      observe: (listener) => {
        assertNotDisposed();
        const index = subscriptions.push(listener) - 1;
        return {
          revoke() {
            delete subscriptions[index];
          }
        };
      },
      dispose: () => {
        assertNotDisposed();
        disposed = true;
        dispose == null ? void 0 : dispose();
      }
    };
  }
  $schedule(task) {
    this.$__scheduler.enqueue(task);
  }
  $tick() {
    this.$__scheduler.flush();
  }
};
export {
  Model,
  debounce,
  effect,
  extend,
  ignore,
  input,
  readonly,
  throttle
};
