import { Task } from './scheduler';
import { type Subscription, type SubscriptionListener } from './subscription';
interface Observable {
    observe(listener: SubscriptionListener): Subscription;
}
export interface Reference<T = unknown> extends Observable {
    get(): T;
    set(value: T): void;
}
export interface Computed<T = unknown> extends Observable {
    get(): T;
    dispose(): void;
}
export interface Extender<T> {
    notify?(value: T): boolean | Promise<boolean>;
    recompute?(): boolean | Promise<boolean>;
    compute?(value: T): T;
    initialize?(target: Computed<T>): void;
}
export declare function ignore(target: object, key: string | symbol): void;
export declare function effect(target: object, key: string | symbol): void;
export declare function throttle(delay: number): (target: object, key: string | symbol) => void;
export declare function debounce(delay: number): (target: object, key: string | symbol) => void;
export declare function readonly(target: object, key: string | symbol): void;
export declare function extend(extender: Extender<unknown>): (target: object, key: string | symbol) => void;
export declare abstract class Model {
    $silent<T>(target: T): T;
    $ref<T>(target: T): Reference<T>;
    $observe(target: unknown, listener: () => void): Subscription;
    /**
     * Creates an computed _value_ which recomputes when a dependency in the current model changes.
     *
     * @example
     * ```typescript
     * // This will recompute when the value of `x` or `y` changes.
     * this.$computed(() => {
     * 	return this.x + this.y
     * })
     * ```
     */
    $computed<T>(get: () => T): Computed<T>;
    $schedule(task: Task): void;
    $tick(): void;
    constructor();
}
export {};
